@@ -1,10 +1,9 @@
 # internals.py - many internal helpers
 
-import datetime
+from datetime import timezone, datetime, timedelta
 from contextlib import closing
 import mimetypes
 import threading
-import six
 import uuid
 import dateutil.parser
 import time
@@ -14,19 +13,22 @@ import re
 import collections
 import base64
 import os
+import inspect
 
-from .utility import AttribAccessDict, AttribAccessList, parse_version_string
-from .errors import MastodonNetworkError, MastodonIllegalArgumentError, MastodonRatelimitError, MastodonNotFoundError, \
+from mastodon.versions import parse_version_string
+from mastodon.errors import MastodonNetworkError, MastodonIllegalArgumentError, MastodonRatelimitError, MastodonNotFoundError, \
                     MastodonUnauthorizedError, MastodonInternalServerError, MastodonBadGatewayError, MastodonServiceUnavailableError, \
                     MastodonGatewayTimeoutError, MastodonServerError, MastodonAPIError, MastodonMalformedEventError
-from .compat import urlparse, magic, PurePath
-from .defaults import _DEFAULT_STREAM_TIMEOUT, _DEFAULT_STREAM_RECONNECT_WAIT_SEC
+from mastodon.compat import urlparse, magic, PurePath, Path
+from mastodon.defaults import _DEFAULT_STREAM_TIMEOUT, _DEFAULT_STREAM_RECONNECT_WAIT_SEC
+from mastodon.return_types import AttribAccessDict, PaginatableList, try_cast_recurse
+from mastodon.return_types import *
 
 ###
 # Internal helpers, dragons probably
 ###
-class Mastodon(): 
-    def __datetime_to_epoch(self, date_time):
+class Mastodon():
+    def __datetime_to_epoch(self, date_time: datetime) -> float:
         """
         Converts a python datetime to unix epoch, accounting for
         time zones and such.
@@ -34,10 +36,9 @@ class Mastodon():
         Assumes UTC if timezone is not given.
         """
         if date_time.tzinfo is None:
-            date_time = date_time.replace(tzinfo=datetime.timezone.utc)
+            date_time = date_time.replace(tzinfo=timezone.utc)
         return date_time.timestamp()
 
-
     def __get_logged_in_id(self):
         """
         Fetch the logged in user's ID, with caching. ID is reset on calls to log_in.
@@ -47,98 +48,53 @@ class Mastodon():
         return self.__logged_in_id
 
     @staticmethod
-    def __json_allow_dict_attrs(json_object):
-        """
-        Makes it possible to use attribute notation to access a dicts
-        elements, while still allowing the dict to act as a dict.
-        """
-        if isinstance(json_object, dict):
-            return AttribAccessDict(json_object)
-        return json_object
-
-    @staticmethod
-    def __json_date_parse(json_object):
-        """
-        Parse dates in certain known json fields, if possible.
-        """
-        known_date_fields = ["created_at", "week", "day", "expires_at", "scheduled_at",
-                                "updated_at", "last_status_at", "starts_at", "ends_at", "published_at", "edited_at", "date", "period"]
-        mark_delete = []
-        for k, v in json_object.items():
-            if k in known_date_fields:
-                if v is not None:
-                    try:
-                        if isinstance(v, int):
-                            json_object[k] = datetime.datetime.fromtimestamp(v, datetime.timezone.utc)
-                        else:
-                            json_object[k] = dateutil.parser.parse(v)
-                    except:
-                        # When we can't parse a date, we just leave the field out
-                        mark_delete.append(k)
-        # Two step process because otherwise python gets very upset
-        for k in mark_delete:
-            del json_object[k]                        
-        return json_object
-
-    @staticmethod
-    def __json_truefalse_parse(json_object):
-        """
-        Parse 'True' / 'False' strings in certain known fields
-        """
-        for key in ('follow', 'favourite', 'reblog', 'mention', 'confirmed', 'suspended', 'silenced', 'disabled', 'approved', 'all_day'):
-            if (key in json_object and isinstance(json_object[key], six.text_type)):
-                if json_object[key].lower() == 'true':
-                    json_object[key] = True
-                if json_object[key].lower() == 'false':
-                    json_object[key] = False
-        return json_object
-
-    @staticmethod
-    def __json_strnum_to_bignum(json_object):
-        """
-        Converts json string numerals to native python bignums.
-        """
-        for key in ('id', 'week', 'in_reply_to_id', 'in_reply_to_account_id', 'logins', 'registrations', 'statuses', 
-                    'day', 'last_read_id', 'value', 'frequency', 'rate', 'invited_by_account_id', 'count'):
-            if (key in json_object and isinstance(json_object[key], six.text_type)):
-                try:
-                    json_object[key] = int(json_object[key])
-                except ValueError:
-                    pass
-
-        return json_object
-
-    @staticmethod
-    def __json_hooks(json_object):
-        """
-        All the json hooks. Used in request parsing.
-        """
-        json_object = Mastodon.__json_strnum_to_bignum(json_object)
-        json_object = Mastodon.__json_date_parse(json_object)
-        json_object = Mastodon.__json_truefalse_parse(json_object)
-        json_object = Mastodon.__json_allow_dict_attrs(json_object)
-        return json_object
-
-    @staticmethod
-    def __consistent_isoformat_utc(datetime_val):
+    def __consistent_isoformat_utc(datetime_val: datetime) -> str:
         """
         Function that does what isoformat does but it actually does the same
         every time instead of randomly doing different things on some systems
         and also it represents that time as the equivalent UTC time.
         """
-        isotime = datetime_val.astimezone(datetime.timezone.utc).strftime("%Y-%m-%dT%H:%M:%S%z")
+        isotime = datetime_val.astimezone(timezone.utc).strftime("%Y-%m-%dT%H:%M:%S%z")
         if isotime[-2] != ":":
             isotime = isotime[:-2] + ":" + isotime[-2:]
         return isotime
 
-    def __api_request(self, method, endpoint, params={}, files={}, headers={}, access_token_override=None, base_url_override=None, 
-                        do_ratelimiting=True, use_json=False, parse=True, return_response_object=False, skip_error_check=False, lang_override=None):
+    def __try_cast_to_type(self, value, override_type = None):
+        """
+        Tries to cast a value to the type of the function two levels up in the call stack.
+        Tries to cast to AttribAccessDict if it doesn't know what to cast to.
+
+        This is used internally inside of __api_request.
+        """
+        try:
+            if override_type is None:
+                # Find type of function two frames up
+                caller_frame = inspect.currentframe().f_back.f_back
+                caller_function = caller_frame.f_code
+                caller_func_name = caller_function.co_name
+                func_obj = getattr(self, caller_func_name)
+
+                # Very carefully try to find what we need to cast to
+                return_type = AttribAccessDict
+                if func_obj is not None:
+                    return_type = func_obj.__annotations__.get('return', AttribAccessDict)
+            else:
+                return_type = override_type
+        except:
+            return_type = AttribAccessDict
+        return try_cast_recurse(return_type, value)
+
+    def __api_request(self, method, endpoint, params={}, files={}, headers={}, access_token_override=None, base_url_override=None,
+                        do_ratelimiting=True, use_json=False, parse=True, return_response_object=False, skip_error_check=False, lang_override=None, override_type=None,
+                        force_pagination=False):
         """
         Internal API request helper.
+
+        Does a large amount of different things that I should document one day, but not today.
         """
         response = None
         remaining_wait = 0
-        
+
         # Add language to params if not None
         lang = self.lang
         if lang_override is not None:
@@ -182,11 +138,10 @@ class Mastodon():
             base_url = base_url_override
 
         if self.debug_requests:
-            print('Mastodon: Request to endpoint "' + base_url +
-                    endpoint + '" using method "' + method + '".')
-            print('Parameters: ' + str(params))
-            print('Headers: ' + str(headers))
-            print('Files: ' + str(files))
+            print(f'Mastodon: Request to endpoint "{base_url}{endpoint}" using method "{method}".')
+            print(f'Parameters: {params}')
+            print(f'Headers: {headers}')
+            print(f'Files: {files}')
 
         # Make request
         request_complete = False
@@ -204,8 +159,12 @@ class Mastodon():
                     kwargs['data'] = params
 
                 response_object = self.session.request(method, base_url + endpoint, **kwargs)
+                if self.debug_requests:
+                    print(f'Mastodon: Request URL: {response_object.request.url}')
+                    print(f'Mastodon: Request body: {response_object.request.body}')
+                    print(f'Mastodon: Response body: {response_object.text}')
             except Exception as e:
-                raise MastodonNetworkError("Could not complete request: %s" % e)
+                raise MastodonNetworkError(f"Could not complete request: {e}")
 
             if response_object is None:
                 raise MastodonIllegalArgumentError("Illegal request.")
@@ -219,8 +178,7 @@ class Mastodon():
 
                 # For gotosocial, we need an int representation, but for non-ints this would crash
                 try:
-                    ratelimit_intrep = str(
-                        int(response_object.headers['X-RateLimit-Reset']))
+                    ratelimit_intrep = str(int(response_object.headers['X-RateLimit-Reset']))
                 except:
                     ratelimit_intrep = None
 
@@ -240,17 +198,17 @@ class Mastodon():
                         self.ratelimit_reset += server_time_diff
                         self.ratelimit_lastcall = time.time()
                 except Exception as e:
-                    raise MastodonRatelimitError("Rate limit time calculations failed: %s" % e)
+                    raise MastodonRatelimitError(f"Rate limit time calculations failed: {e}")
 
             # Handle response
             if self.debug_requests:
-                print('Mastodon: Response received with code ' + str(response_object.status_code) + '.')
-                print('response headers: ' + str(response_object.headers))
-                print('Response text content: ' + str(response_object.text))
+                print(f'Mastodon: Response received with code {response_object.status_code}.')
+                print(f'response headers: {response_object.headers}')
+                print(f'Response text content: {response_object.text}')
 
             if not response_object.ok:
                 try:
-                    response = response_object.json(object_hook=self.__json_hooks)
+                    response = self.__try_cast_to_type(response_object.json(), override_type = override_type) # TODO actually cast to an error type
                     if isinstance(response, dict) and 'error' in response:
                         error_msg = response['error']
                     elif isinstance(response, str):
@@ -272,7 +230,7 @@ class Mastodon():
                             time.sleep(to_next)
                             request_complete = False
                             continue
-                
+
                 if not skip_error_check:
                     if response_object.status_code == 404:
                         ex_type = MastodonNotFoundError
@@ -303,22 +261,24 @@ class Mastodon():
 
             if parse:
                 try:
-                    response = response_object.json(object_hook=self.__json_hooks)
-                except:
+                    # The new parsing is very basic, type conversion happens later,
+                    # within the new type system. This should be overall more robust.
+                    response = response_object.json()
+                except Exception as e:
                     raise MastodonAPIError(
-                        "Could not parse response as JSON, response code was %s, "
-                        "bad json content was '%s'" % (response_object.status_code,
-                                                        response_object.content))
+                        f"Could not parse response as JSON, response code was {response_object.status_code}, "
+                        f"bad json content was {response_object.content!r}.",
+                        f"Exception was: {e}"
+                    )
+                response = self.__try_cast_to_type(response, override_type = override_type)
             else:
                 response = response_object.content
 
             # Parse link headers
-            if isinstance(response, list) and \
-                    'Link' in response_object.headers and \
-                    response_object.headers['Link'] != "":
-                response = AttribAccessList(response)
-                tmp_urls = requests.utils.parse_header_links(
-                    response_object.headers['Link'].rstrip('>').replace('>,<', ',<'))
+            if (isinstance(response, list) or force_pagination) and 'Link' in response_object.headers and response_object.headers['Link'] != "":
+                if not isinstance(response, PaginatableList) and not force_pagination:
+                    response = PaginatableList(response)
+                tmp_urls = requests.utils.parse_header_links(response_object.headers['Link'].rstrip('>').replace('>,<', ',<'))
                 for url in tmp_urls:
                     if 'rel' not in url:
                         continue
@@ -343,18 +303,12 @@ class Mastodon():
                                 del next_params['min_id']
                             response._pagination_next = next_params
 
-                            # Maybe other API users rely on the pagination info in the last item
-                            # Will be removed in future
-                            if isinstance(response[-1], AttribAccessDict):
-                                response[-1]._pagination_next = next_params
-
                     if url['rel'] == 'prev':
                         # Be paranoid and extract since_id or min_id specifically
                         prev_url = url['url']
 
                         # Old and busted (pre-2.6.0): since_id pagination
-                        matchgroups = re.search(
-                            r"[?&]since_id=([^&]+)", prev_url)
+                        matchgroups = re.search(r"[?&]since_id=([^&]+)", prev_url)
                         if matchgroups:
                             prev_params = copy.deepcopy(params)
                             prev_params['_pagination_method'] = method
@@ -368,14 +322,8 @@ class Mastodon():
                                 del prev_params['max_id']
                             response._pagination_prev = prev_params
 
-                            # Maybe other API users rely on the pagination info in the first item
-                            # Will be removed in future
-                            if isinstance(response[0], AttribAccessDict):
-                                response[0]._pagination_prev = prev_params
-
                         # New and fantastico (post-2.6.0): min_id pagination
-                        matchgroups = re.search(
-                            r"[?&]min_id=([^&]+)", prev_url)
+                        matchgroups = re.search(r"[?&]min_id=([^&]+)", prev_url)
                         if matchgroups:
                             prev_params = copy.deepcopy(params)
                             prev_params['_pagination_method'] = method
@@ -388,22 +336,31 @@ class Mastodon():
                             if "max_id" in prev_params:
                                 del prev_params['max_id']
                             response._pagination_prev = prev_params
-
-                            # Maybe other API users rely on the pagination info in the first item
-                            # Will be removed in future
-                            if isinstance(response[0], AttribAccessDict):
-                                response[0]._pagination_prev = prev_params
-
+        
         return response
 
-    def __get_streaming_base(self):
+    def __get_streaming_base(self) -> str:
         """
         Internal streaming API helper.
 
         Returns the correct URL for the streaming API.
         """
-        instance = self.instance()
-        if "streaming_api" in instance["urls"] and instance["urls"]["streaming_api"] != self.api_base_url:
+        if self.__streaming_base is not None:
+            return self.__streaming_base
+        
+        # Try to support implementations that have no v1 endpoint (Sharkey does this)
+        streaming_api_url = None
+        try:
+            instance = self.__instance()
+            if  "streaming_api" in instance["urls"]:
+                streaming_api_url = instance["urls"]["streaming_api"]
+        except:
+            try:
+                streaming_api_url = self.instance_v2().configuration.urls.streaming
+            except:
+                pass
+
+        if not streaming_api_url is None and streaming_api_url != self.api_base_url:
             # This is probably a websockets URL, which is really for the browser, but requests can't handle it
             # So we do this below to turn it into an HTTPS or HTTP URL
             parse = urlparse(instance["urls"]["streaming_api"])
@@ -413,10 +370,11 @@ class Mastodon():
                 url = "http://" + parse.netloc
             else:
                 raise MastodonAPIError(
-                    "Could not parse streaming api location returned from server: {}.".format(
-                        instance["urls"]["streaming_api"]))
+                    f"Could not parse streaming api location returned from server: {instance['urls']['streaming_api']}."
+                )
         else:
             url = self.api_base_url
+        assert not url is None
         return url
 
     def __stream(self, endpoint, listener, params={}, run_async=False, timeout=_DEFAULT_STREAM_TIMEOUT, reconnect_async=False, reconnect_async_wait_sec=_DEFAULT_STREAM_RECONNECT_WAIT_SEC):
@@ -436,16 +394,14 @@ class Mastodon():
 
         # Connect function (called and then potentially passed to async handler)
         def connect_func():
-            headers = {"Authorization": "Bearer " +
-                        self.access_token} if self.access_token else {}
+            headers = {"Authorization": "Bearer " + self.access_token} if self.access_token else {}
             if self.user_agent:
                 headers['User-Agent'] = self.user_agent
             connection = self.session.get(url + endpoint, headers=headers, data=params, stream=True,
                                             timeout=(self.request_timeout, timeout))
 
             if connection.status_code != 200:
-                raise MastodonNetworkError(
-                    "Could not connect to streaming server: %s" % connection.reason)
+                raise MastodonNetworkError(f"Could not connect to streaming server: {connection.reason}")
             return connection
         connection = None
 
@@ -543,7 +499,7 @@ class Mastodon():
             with closing(connection) as r:
                 listener.handle_stream(r)
 
-    def __generate_params(self, params, exclude=[]):
+    def __generate_params(self, params, exclude=[], dateconv=False, for_json=False):
         """
         Internal named-parameters-to-dict helper.
 
@@ -566,15 +522,23 @@ class Mastodon():
             if params[key] is None or key in exclude:
                 del params[key]
 
-        param_keys = list(params.keys())
+        if not for_json:
+            param_keys = list(params.keys())
+            for key in param_keys:
+                if isinstance(params[key], list):
+                    params[key + "[]"] = params[key]
+                    del params[key]
+
+        # Unpack min/max/since_id fields, since that is a very common operation
+        # and we basically always want it
         for key in param_keys:
-            if isinstance(params[key], list):
-                params[key + "[]"] = params[key]
-                del params[key]
+            if key in ['min_id', 'max_id', 'since_id']:
+                print("Unpacking", key)
+                params[key] = self.__unpack_id(params[key], dateconv = dateconv, listify = False)
 
         return params
 
-    def __unpack_id(self, id, dateconv=False):
+    def __unpack_id(self, id, dateconv = False, listify = False, field = "id"):
         """
         Internal object-to-id converter
 
@@ -584,9 +548,17 @@ class Mastodon():
 
         Also unpacks datetimes to snowflake IDs if requested.
         """
-        if isinstance(id, dict) and "id" in id:
-            id = id["id"]
-        if dateconv and isinstance(id, datetime.datetime):
+        if id is None:
+            return None
+        if not isinstance(id, list) and listify:
+            id = [id]
+        if isinstance(id, list):
+            for i in range(len(id)):
+                id[i] = self.__unpack_id(id[i], dateconv = dateconv, listify = False)
+            return id
+        if isinstance(id, dict) and field in id:
+            id = id[field]
+        if dateconv and isinstance(id, datetime):
             id = (int(id.timestamp()) << 16) * 1000
         return id
 
@@ -601,11 +573,11 @@ class Mastodon():
 
     def __get_token_expired(self):
         """Internal helper for oauth code"""
-        return self._token_expired < datetime.datetime.now()
+        return self._token_expired < datetime.now()
 
     def __set_token_expired(self, value):
         """Internal helper for oauth code"""
-        self._token_expired = datetime.datetime.now() + datetime.timedelta(seconds=value)
+        self._token_expired = datetime.now() + timedelta(seconds=value)
         return
 
     def __get_refresh_token(self):
@@ -627,18 +599,22 @@ class Mastodon():
         return mime_type
 
     def __load_media_file(self, media_file, mime_type=None, file_name=None):
+        """Internal helper to load a media file"""
         if isinstance(media_file, PurePath):
             media_file = str(media_file)
+        if isinstance(media_file, str):
+            try: # Explicitly resolve to canonical for robustness. This can and will fail if Path isn't available because python too old.
+                media_file = str(Path(media_file).resolve())
+            except:
+                pass
         if isinstance(media_file, str) and os.path.isfile(media_file):
             mime_type = self.__guess_type(media_file)
             media_file = open(media_file, 'rb')
-        elif isinstance(media_file, str) and os.path.isfile(media_file):
-            media_file = open(media_file, 'rb')
         if mime_type is None:
             raise MastodonIllegalArgumentError('Could not determine mime type or data passed directly without mime type.')
         if file_name is None:
             random_suffix = uuid.uuid4().hex
-            file_name = "mastodonpyupload_" + str(time.time()) + "_" + str(random_suffix) + mimetypes.guess_extension(mime_type)
+            file_name = f"mastodonpyupload_{time.time()}_{random_suffix}{mimetypes.guess_extension(mime_type)}"
         return (file_name, media_file, mime_type)
 
     @staticmethod
@@ -665,8 +641,11 @@ class Mastodon():
         version_string = version_string.split(" ")[0]
         try:
             # Attempt to split at + and check if the part after parses as a version string, to account for hometown
-            parse_version_string(version_string.split("+")[1])
+            ver_parts = parse_version_string(version_string.split("+")[1])
+            # If the parsed version is less than 1.0, assume it's GoToSocial and return the *first* part
+            if ver_parts[0] < 1:
+                return version_string.split("+")[0]
             return version_string.split("+")[1]
         except:
             # If this fails, assume that if there is a +, what is before that is the masto version (or that there is no +)
-            return version_string.split("+")[0]
\ No newline at end of file
+            return version_string.split("+")[0]
